/*
 * M6 Implementation
 *
 * ==========================(LICENSE BEGIN)============================
 *
 * Copyright (c) 2014  phm
 * Copyright (c) 2014 Girino Vey
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * ===========================(LICENSE END)=============================
 *
 * @author   rafal <rafal.skowronski@cs.put.poznan.pl>
 */



#define  SPH_LITTLE_ENDIAN 1
#define SPH_UPTR sph_u64

typedef unsigned int sph_u32;
typedef int sph_s32;
#ifndef __OPENCL_VERSION__
typedef unsigned long long sph_u64;
typedef long long sph_s64;
#else
typedef unsigned long sph_u64;
typedef long sph_s64;
#endif

#define SPH_64 1
#define SPH_64_TRUE 1

#define SPH_C32(x)    ((sph_u32)(x ## U))
#define SPH_T32(x)    ((x) & SPH_C32(0xFFFFFFFF))
#define SPH_ROTL32(x, n)   SPH_T32(((x) << (n)) | ((x) >> (32 - (n))))
#define SPH_ROTR32(x, n)   SPH_ROTL32(x, (32 - (n)))

#define SPH_C64(x)    ((sph_u64)(x ## UL))
#define SPH_T64(x)    ((x) & SPH_C64(0xFFFFFFFFFFFFFFFF))
#define SPH_ROTL64(x, n)   SPH_T64(((x) << (n)) | ((x) >> (64 - (n))))
#define SPH_ROTR64(x, n)   SPH_ROTL64(x, (64 - (n)))

#define SPH_ECHO_64 1
#define SPH_KECCAK_64 1
#define SPH_JH_64 1
#define SPH_SIMD_NOCOPY 0
#define SPH_KECCAK_NOCOPY 0
#define SPH_COMPACT_BLAKE_64 0
#define SPH_LUFFA_PARALLEL 0
#define SPH_SMALL_FOOTPRINT_GROESTL 0
#define SPH_GROESTL_BIG_ENDIAN 0
#define SPH_CUBEHASH_UNROLL 0
#define SPH_KECCAK_UNROLL   0
#define SPH_HAMSI_EXPAND_BIG 4

#include "blake.cl"
#include "jh.cl"
#include "keccak.cl"
#include "skein.cl"
#include "shavite.cl"
#include "hamsi.cl"


#define SWAP4(x) as_uint(as_uchar4(x).wzyx)
#define SWAP8(x) as_ulong(as_uchar8(x).s76543210)


#define DEC64E(x) SWAP8(x)
    #define DEC64BE(x) SWAP8(*(const __global sph_u64 *) (x));
	    #define DEC64BEC(x) *(const __global sph_u64 *) (x);

typedef union {
    unsigned char h1[64];
    uint h4[16];
    ulong h8[8];
} hash_t;

typedef struct hsd
{
    hash_t cross_h;
	hash_t temp_h;
} h_pair_t;


//we use both static reqd_work_group_size and dynamic local_work_size at kernel dispacth
// to allow for compiler to make some compile-time device-specific optimisations

__attribute__((reqd_work_group_size(WORKSIZE, 1, 1)))
__kernel void blake(__global unsigned char* block, 
__global h_pair_t *hashes,uint isHeader)
{
    size_t gid = get_global_id(0);
	unsigned int gid32=(unsigned int)gid;
	 uint offset  = (uint)(get_global_id(0) - get_global_offset(0));
    __global h_pair_t *h_pair= &(hashes[offset]);
    // blake

    sph_u64 H0 = SPH_C64(0x6A09E667F3BCC908), H1 = SPH_C64(0xBB67AE8584CAA73B);
    sph_u64 H2 = SPH_C64(0x3C6EF372FE94F82B), H3 = SPH_C64(0xA54FF53A5F1D36F1);
    sph_u64 H4 = SPH_C64(0x510E527FADE682D1), H5 = SPH_C64(0x9B05688C2B3E6C1F);
    sph_u64 H6 = SPH_C64(0x1F83D9ABFB41BD6B), H7 = SPH_C64(0x5BE0CD19137E2179);
    sph_u64 S0 = 0, S1 = 0, S2 = 0, S3 = 0;
    sph_u64 T0 = SPH_C64(0xFFFFFFFFFFFFFC00) + (80 << 3), T1 = 0xFFFFFFFFFFFFFFFF;;

    if ((T0 = SPH_T64(T0 + 1024)) < 1024)
    {
        T1 = SPH_T64(T1 + 1);
    }
    sph_u64 M0, M1, M2, M3, M4, M5, M6, M7;
    sph_u64 M8, M9, MA, MB, MC, MD, ME, MF;
    sph_u64 V0, V1, V2, V3, V4, V5, V6, V7;
    sph_u64 V8, V9, VA, VB, VC, VD, VE, VF;
	if(false)
	{
    M0 =  DEC64BE(block +   0);
    M1 =  DEC64BE(block +   8);
    M2 =  DEC64BE(block +  16);
    M3 =  DEC64BE(block +  24);
    M4 =  DEC64BE(block +  32);
    M5 =  DEC64BE(block +  40);
    M6 =  DEC64BE(block +  48);
    M7 =  DEC64BE(block +  56);
    M8 =  DEC64BE(block +  64);
    M9 =  DEC64BEC(block +  72);
    M9 &= 0xFFFFFFFF00000000;
    M9 ^= gid32;
	M9 =  SWAP8(M9);
    MA = 0x8000000000000000;
    MB = 0;
    MC = 0;
    MD = 1;
    ME = 0;
    MF = 0x280;
	}
	else{
	//used for a regular PoW ex. for Identity Token PoW
	//here a 512 Keccak Hash of input data is processed by Ultimium.
	//a 32-bit nonce is appended to the end.
	
	//512-bit keccach hash:
	M0 =  DEC64BE(block +   0);
    M1 =  DEC64BE(block +   8);
    M2 =  DEC64BE(block +  16);
    M3 =  DEC64BEC(block +  24);
	M3 &= 0xFFFFFFFF00000000;
    M3 ^= gid32;
	M3 =  SWAP8(M3);
    M4 =  DEC64BE(block +  32);
    M5 =  DEC64BE(block +  40);
    M6 =  DEC64BE(block +  48);
    M7 =  DEC64BE(block +  56);
    M8 =  DEC64BE(block +  64);
    M9 =  DEC64BE(block +  72);
    
    MA = 0x8000000000000000;
    MB = 0;
    MC = 0;
    MD = 1;
    ME = 0;
    MF = 0x280;
	}


    COMPRESS64;

	h_pair->temp_h.h8[0] = SWAP8(H0), 			h_pair->cross_h.h8[0] = SWAP8(H0);
	h_pair->temp_h.h8[1] = SWAP8(H1), 			h_pair->cross_h.h8[1] = SWAP8(H1);
	h_pair->temp_h.h8[2] = SWAP8(H2), 		    h_pair->cross_h.h8[2] = SWAP8(H2);
	h_pair->temp_h.h8[3] = SWAP8(H3),           h_pair->cross_h.h8[3] = SWAP8(H3);
	h_pair->temp_h.h8[4] = SWAP8(H4);
	h_pair->temp_h.h8[5] = SWAP8(H5);
	h_pair->temp_h.h8[6] = SWAP8(H6);
	h_pair->temp_h.h8[7] = SWAP8(H7);

	
	
	

    barrier(CLK_GLOBAL_MEM_FENCE);
}

//2
__attribute__((reqd_work_group_size(WORKSIZE, 1, 1)))
__kernel void keccak(__global h_pair_t* hashes )
{
 uint offset  = (uint)(get_global_id(0) - get_global_offset(0));
    __global h_pair_t *h_pair= &(hashes[offset]);

    // keccak

    sph_u64 a00 = 0, a01 = 0, a02 = 0, a03 = 0, a04 = 0;
    sph_u64 a10 = 0, a11 = 0, a12 = 0, a13 = 0, a14 = 0; 
    sph_u64 a20 = 0, a21 = 0, a22 = 0, a23 = 0, a24 = 0;
    sph_u64 a30 = 0, a31 = 0, a32 = 0, a33 = 0, a34 = 0;
    sph_u64 a40 = 0, a41 = 0, a42 = 0, a43 = 0, a44 = 0;

    a10 = SPH_C64(0xFFFFFFFFFFFFFFFF);
    a20 = SPH_C64(0xFFFFFFFFFFFFFFFF);
    a31 = SPH_C64(0xFFFFFFFFFFFFFFFF);
    a22 = SPH_C64(0xFFFFFFFFFFFFFFFF);
    a23 = SPH_C64(0xFFFFFFFFFFFFFFFF);
    a04 = SPH_C64(0xFFFFFFFFFFFFFFFF);

    a00 ^= h_pair->temp_h.h8[0];
    a10 ^= h_pair->temp_h.h8[1];
    a20 ^= h_pair->temp_h.h8[2];
    a30 ^= h_pair->temp_h.h8[3];
    a40 ^= h_pair->temp_h.h8[4];
    a01 ^= h_pair->temp_h.h8[5];
    a11 ^= h_pair->temp_h.h8[6];
    a21 ^= h_pair->temp_h.h8[7];
    a31 ^= 0x8000000000000001;
    KECCAK_F_1600;
    // Finalize the "lane complement"
    a10 = ~a10;
    a20 = ~a20;

	h_pair->temp_h.h8[0] = a00; 
	h_pair->temp_h.h8[1] = a10;
	h_pair->temp_h.h8[2] = a20;
	h_pair->temp_h.h8[3] = a30;
	h_pair->temp_h.h8[4] = a40; h_pair->cross_h.h8[4] = a40;
	h_pair->temp_h.h8[5] = a01; h_pair->cross_h.h8[5] = a01;
	h_pair->temp_h.h8[6] = a11; h_pair->cross_h.h8[6] = a11;
	h_pair->temp_h.h8[7] = a21; h_pair->cross_h.h8[7] = a21;

	
	
	
    barrier(CLK_GLOBAL_MEM_FENCE);
}

//3
__attribute__((reqd_work_group_size(WORKSIZE, 1, 1)))
__kernel void hamsi(__global h_pair_t *hashes)
{

 uint offset  = (uint)(get_global_id(0) - get_global_offset(0));
    __global h_pair_t *h_pair= &(hashes[offset]);
	
	
    hash_t hash;



    sph_u32 c0 = HAMSI_IV512[0], c1 = HAMSI_IV512[1], c2 = HAMSI_IV512[2], c3 = HAMSI_IV512[3];
    sph_u32 c4 = HAMSI_IV512[4], c5 = HAMSI_IV512[5], c6 = HAMSI_IV512[6], c7 = HAMSI_IV512[7];
    sph_u32 c8 = HAMSI_IV512[8], c9 = HAMSI_IV512[9], cA = HAMSI_IV512[10], cB = HAMSI_IV512[11];
    sph_u32 cC = HAMSI_IV512[12], cD = HAMSI_IV512[13], cE = HAMSI_IV512[14], cF = HAMSI_IV512[15];
    sph_u32 m0, m1, m2, m3, m4, m5, m6, m7;
    sph_u32 m8, m9, mA, mB, mC, mD, mE, mF;
    sph_u32 h[16] = { c0, c1, c2, c3, c4, c5, c6, c7, c8, c9, cA, cB, cC, cD, cE, cF };

    for (int i = 0; i < 8; i++) {
        hash.h8[i] = h_pair->cross_h.h8[i];
    }

#define buf(u) hash.h1[i + u]
    for(int i = 0; i < 64; i += 8) {
        INPUT_BIG;
        P_BIG;
        T_BIG;
    }
#undef buf
#define buf(u) (u == 0 ? 0x80 : 0)
    INPUT_BIG;
    P_BIG;
    T_BIG;
#undef buf
#define buf(u) (u == 6 ? 2 : 0)
    INPUT_BIG;
    PF_BIG;
    T_BIG;

  
		
   for (unsigned u = 0; u < 16; u ++)
        h_pair->temp_h.h4[u] = SWAP4(h[u]);
		
   for (unsigned u = 0; u < 8; u ++)//copy just the first half to cross-temp
        h_pair->cross_h.h4[u] = SWAP4(h[u]);
		
 

    barrier(CLK_GLOBAL_MEM_FENCE);
}

//4
__attribute__((reqd_work_group_size(WORKSIZE, 1, 1)))
__kernel void shavite(__global h_pair_t *hashes)
{
    uint offset  = (uint)(get_global_id(0) - get_global_offset(0));
	
	 __global h_pair_t *h_pair= &(hashes[offset]);
    __local sph_u32 AES0[256], AES1[256], AES2[256], AES3[256];

    uint init = get_local_id(0);
    uint step = get_local_size(0);

    for (int i = init; i < 256; i += step)
    {
        AES0[i] = AES0_C[i];
        AES1[i] = AES1_C[i];
        AES2[i] = AES2_C[i];
        AES3[i] = AES3_C[i];
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    // shavite
    // IV
    sph_u32 h0 = SPH_C32(0x72FCCDD8), h1 = SPH_C32(0x79CA4727), h2 = SPH_C32(0x128A077B), h3 = SPH_C32(0x40D55AEC);
    sph_u32 h4 = SPH_C32(0xD1901A06), h5 = SPH_C32(0x430AE307), h6 = SPH_C32(0xB29F5CD1), h7 = SPH_C32(0xDF07FBFC);
    sph_u32 h8 = SPH_C32(0x8E45D73D), h9 = SPH_C32(0x681AB538), hA = SPH_C32(0xBDE86578), hB = SPH_C32(0xDD577E47);
    sph_u32 hC = SPH_C32(0xE275EADE), hD = SPH_C32(0x502D9FCD), hE = SPH_C32(0xB9357178), hF = SPH_C32(0x022A4B9A);

    // state
    sph_u32 rk00, rk01, rk02, rk03, rk04, rk05, rk06, rk07;
    sph_u32 rk08, rk09, rk0A, rk0B, rk0C, rk0D, rk0E, rk0F;
    sph_u32 rk10, rk11, rk12, rk13, rk14, rk15, rk16, rk17;
    sph_u32 rk18, rk19, rk1A, rk1B, rk1C, rk1D, rk1E, rk1F;

    sph_u32 sc_count0 = (64 << 3), sc_count1 = 0, sc_count2 = 0, sc_count3 = 0;

    rk00 = h_pair->temp_h.h4[0];
    rk01 = h_pair->temp_h.h4[1];
    rk02 = h_pair->temp_h.h4[2];
    rk03 = h_pair->temp_h.h4[3];
    rk04 = h_pair->temp_h.h4[4];
    rk05 = h_pair->temp_h.h4[5];
    rk06 = h_pair->temp_h.h4[6];
    rk07 = h_pair->temp_h.h4[7];
    rk08 = h_pair->temp_h.h4[8];
    rk09 = h_pair->temp_h.h4[9];
    rk0A = h_pair->temp_h.h4[10];
    rk0B = h_pair->temp_h.h4[11];
    rk0C = h_pair->temp_h.h4[12];
    rk0D = h_pair->temp_h.h4[13];
    rk0E = h_pair->temp_h.h4[14];
    rk0F = h_pair->temp_h.h4[15];
    rk10 = 0x80;
    rk11 = rk12 = rk13 = rk14 = rk15 = rk16 = rk17 = rk18 = rk19 = rk1A = 0;
    rk1B = 0x2000000;
    rk1C = rk1D = rk1E = 0;
    rk1F = 0x2000000;

    c512(buf);

	//remove below? only second half to crosstemp needed
    h_pair->temp_h.h4[0] = h0;
    h_pair->temp_h.h4[1] = h1;
    h_pair->temp_h.h4[2] = h2;
    h_pair->temp_h.h4[3] = h3;
    h_pair->temp_h.h4[4] = h4;
    h_pair->temp_h.h4[5] = h5;
    h_pair->temp_h.h4[6] = h6;
    h_pair->temp_h.h4[7] = h7;
	
    h_pair->temp_h.h4[8] = h8, h_pair->cross_h.h4[8] = h8;
    h_pair->temp_h.h4[9] = h9, h_pair->cross_h.h4[9] = h9;
    h_pair->temp_h.h4[10] = hA, h_pair->cross_h.h4[10] = hA;
    h_pair->temp_h.h4[11] = hB, h_pair->cross_h.h4[11] = hB;
    h_pair->temp_h.h4[12] = hC, h_pair->cross_h.h4[12] = hC;
    h_pair->temp_h.h4[13] = hD, h_pair->cross_h.h4[13] = hD;
    h_pair->temp_h.h4[14] = hE, h_pair->cross_h.h4[14] = hE;
    h_pair->temp_h.h4[15] = hF, h_pair->cross_h.h4[15] = hF;

	
	
   
 
	
    barrier(CLK_GLOBAL_MEM_FENCE);
}


//5
__attribute__((reqd_work_group_size(WORKSIZE, 1, 1)))
__kernel void jh(__global h_pair_t *hashes)
{
       uint offset  = (uint)(get_global_id(0) - get_global_offset(0));
    __global h_pair_t *h_pair= &(hashes[offset]);

   // jh

    sph_u64 h0h = C64e(0x6fd14b963e00aa17), h0l = C64e(0x636a2e057a15d543), h1h = C64e(0x8a225e8d0c97ef0b), h1l = C64e(0xe9341259f2b3c361), h2h = C64e(0x891da0c1536f801e), h2l = C64e(0x2aa9056bea2b6d80), h3h = C64e(0x588eccdb2075baa6), h3l = C64e(0xa90f3a76baf83bf7);
    sph_u64 h4h = C64e(0x0169e60541e34a69), h4l = C64e(0x46b58a8e2e6fe65a), h5h = C64e(0x1047a7d0c1843c24), h5l = C64e(0x3b6e71b12d5ac199), h6h = C64e(0xcf57f6ec9db1f856), h6l = C64e(0xa706887c5716b156), h7h = C64e(0xe3c2fcdfe68517fb), h7l = C64e(0x545a4678cc8cdd4b);
    sph_u64 tmp;
	
    for(int i = 0; i < 2; i++)
    {
        if (i == 0) {
		
            h0h ^= h_pair->cross_h.h8[0];
			
            h0l ^= h_pair->cross_h.h8[1];
            h1h ^= h_pair->cross_h.h8[2];
            h1l ^= h_pair->cross_h.h8[3];
			
            h2h ^= h_pair->cross_h.h8[4];
            h2l ^= h_pair->cross_h.h8[5];
            h3h ^= h_pair->cross_h.h8[6];
            h3l ^= h_pair->cross_h.h8[7];
        } else if(i == 1) {
            h4h ^= h_pair->cross_h.h8[0];
            h4l ^= h_pair->cross_h.h8[1];
            h5h ^= h_pair->cross_h.h8[2];
            h5l ^= h_pair->cross_h.h8[3];
			
            h6h ^= h_pair->cross_h.h8[4];
            h6l ^= h_pair->cross_h.h8[5];
            h7h ^= h_pair->cross_h.h8[6];
            h7l ^= h_pair->cross_h.h8[7];
        
            h0h ^= 0x80;
            h3l ^= 0x2000000000000;
        }
        E8;
    }
    h4h ^= 0x80;
    h7l ^= 0x2000000000000;

    h_pair->temp_h.h8[0] = h4h;
    h_pair->temp_h.h8[1] = h4l;
    h_pair->temp_h.h8[2] = h5h;
    h_pair->temp_h.h8[3] = h5l;
    h_pair->temp_h.h8[4] = h6h;
    h_pair->temp_h.h8[5] = h6l;
    h_pair->temp_h.h8[6] = h7h;
    h_pair->temp_h.h8[7] = h7l;

	

	
    barrier(CLK_GLOBAL_MEM_FENCE);
}

__attribute__((reqd_work_group_size(WORKSIZE, 1, 1)))
__kernel void skein(__global h_pair_t *hashes,
 __global uint* output,
 const ulong target, 
 const uint offset,
__global ulong* test)
{
    uint off  = (uint)(get_global_id(0) - get_global_offset(0));
    __global h_pair_t *h_pair= &(hashes[off]);
    // skein

    sph_u64 h0 = SPH_C64(0x4903ADFF749C51CE), h1 = SPH_C64(0x0D95DE399746DF03), h2 = SPH_C64(0x8FD1934127C79BCE), h3 = SPH_C64(0x9A255629FF352CB1), h4 = SPH_C64(0x5DB62599DF6CA7B0), h5 = SPH_C64(0xEABE394CA9D5C3F4), h6 = SPH_C64(0x991112C71A75B523), h7 = SPH_C64(0xAE18A40B660FCC33);
    sph_u64 m0, m1, m2, m3, m4, m5, m6, m7;
    sph_u64 bcount = 0;

    m0 = h_pair->temp_h.h8[0];
    m1 = h_pair->temp_h.h8[1];
    m2 = h_pair->temp_h.h8[2];
    m3 = h_pair->temp_h.h8[3];
    m4 = h_pair->temp_h.h8[4];
    m5 = h_pair->temp_h.h8[5];
    m6 = h_pair->temp_h.h8[6];
    m7 = h_pair->temp_h.h8[7];
    UBI_BIG(480, 64);
    bcount = 0;
    m0 = m1 = m2 = m3 = m4 = m5 = m6 = m7 = 0;
    UBI_BIG(510, 8);
    h_pair->temp_h.h8[0] = h0;
    h_pair->temp_h.h8[1] = h1;
    h_pair->temp_h.h8[2] = h2;
    h_pair->temp_h.h8[3] = h3;
    h_pair->temp_h.h8[4] = h4;
    h_pair->temp_h.h8[5] = h5;
    h_pair->temp_h.h8[6] = h6;
    h_pair->temp_h.h8[7] = h7;
 
    bool result = ((ulong)(h_pair->temp_h.h8[3]) <= (ulong)target);
   
    if (result)
	{
	uint i;
	output[i= atomic_inc(output+0xFF)] = (uint)get_global_id(0);
	test[i] = h_pair->temp_h.h8[3];
	}

	
    barrier(CLK_GLOBAL_MEM_FENCE);
	
}



